from dataclasses import dataclass, field
from typing import assert_never

from little_a2s.events import (
    ClientEvent,
    ClientEventChallenge,
    ClientEventGoldsourceInfo,
    ClientEventInfo,
    ClientEventPlayers,
    ClientEventRules,
)
from little_a2s.headers import (
    Compression as Compression,
    Header as Header,
    HeaderType as HeaderType,
    MultiGoldsourceHeader as MultiGoldsourceHeader,
    MultiHeader as MultiHeader,
    SimpleHeader as SimpleHeader,
)
from little_a2s.packets import (
    ClientPacket,
    ClientPacketInfo,
    ClientPacketPlayers,
    ClientPacketRules,
)
from little_a2s.reader import Reader

MultiAnyHeader = MultiHeader | MultiGoldsourceHeader


@dataclass(kw_only=True)
class MultiPartResponse:
    """A multi-part response waiting to be assembled."""

    id: int
    current: int
    total: int
    compressed: Compression | None
    payloads: dict[int, bytes] = field(default_factory=dict)

    def add(self, header: MultiAnyHeader, payload: bytes) -> None:
        if header.id != self.id:
            raise ValueError(f"Expected ID {self.id!r}, got {header.id!r}")
        elif header.total != self.total:
            raise ValueError(f"Expected total {self.total!r}, got {header.total!r}")
        elif not 0 <= header.current < self.total:
            raise ValueError(
                f"Expected current in range [0, {self.total!r}), got {header.current!r}"
            )
        elif header.current in self.payloads:
            raise ValueError(f"Duplicate current index {header.current!r}")

        self.payloads[header.current] = payload

    def assemble(self) -> bytes | None:
        if len(self.payloads) < self.total:
            return

        payload = b"".join(self.payloads[i] for i in range(self.total))
        if not self.compressed:
            return payload

        return self._decompress_payload(payload)

    def _decompress_payload(self, payload: bytes) -> bytes:
        import bz2
        import zlib

        assert self.compressed is not None
        size = self.compressed.size
        checksum = self.compressed.checksum

        payload = bz2.decompress(payload)

        if len(payload) != size:
            raise ValueError(f"Expected payload size {size}, got {len(payload)}")
        elif zlib.crc32(payload) != checksum:
            raise ValueError(f"Expected checksum {checksum:X}, got {payload:X}")

        return payload


class A2SClientProtocol:
    """Implements the client-side protocol for A2S.

    This follows the Source format. For the Goldsource equivalent,
    see :class:`A2SGoldsourceClientProtocol`.

    :param challenge:
        The challenge sequence to use for requests.
        This can change dynamically when receiving S2C_CHALLENGE responses.

    """

    _events: list[ClientEvent]
    _to_send: list[ClientPacket]
    _responses: dict[int, MultiPartResponse]

    def __init__(self, *, challenge: int = -1) -> None:
        self.challenge = challenge

        self._events = []
        self._to_send = []
        self._responses = {}

    def __repr__(self) -> str:
        type_ = type(self).__name__
        challenge = self.challenge
        events = len(self._events)
        packets = len(self._to_send)
        return f"<{type_} {challenge=}, {events} events, {packets} packets to send>"

    def receive_datagram(self, data: bytes) -> None:
        """Process a packet from the server.

        :raises ValueError: The data is malformed.

        """
        reader = Reader(data)

        try:
            header = self._parse_header(reader)
            payload = self._parse_payload(reader, header)
        except EOFError as e:
            raise ValueError("Received incomplete data") from e

        if payload is not None:
            self._handle_payload(payload)

    def events_received(self) -> list[ClientEvent]:
        """Return a list of events received since this last call."""
        current_events, self._events = self._events, []
        return current_events

    def packets_to_send(self) -> list[ClientPacket]:
        """Return a list of packets to send since this last call."""
        current_datagrams, self._to_send = self._to_send, []
        return current_datagrams

    def info(self) -> ClientPacketInfo:
        """Create an A2S_INFO packet to send to the server."""
        return ClientPacketInfo(challenge=self.challenge)

    def players(self) -> ClientPacketPlayers:
        """Create an A2S_PLAYER packet to send to the server."""
        return ClientPacketPlayers(challenge=self.challenge)

    def rules(self) -> ClientPacketRules:
        """Create an A2S_RULES packet to send to the server."""
        return ClientPacketRules(challenge=self.challenge)

    def invalidate_response(self, id: int) -> MultiPartResponse | None:
        """Invalidate an incomplete multi-part response with the given ID.

        This should be invoked after some sort of timeout.

        """
        return self._responses.pop(id, None)

    def _parse_header(self, reader: Reader) -> Header:
        type = HeaderType(reader.read_long())
        if type == HeaderType.SIMPLE:
            return self._parse_simple_header(reader)
        elif type == HeaderType.MULTI:
            return self._parse_multi_header(reader)
        else:
            assert_never(type)

    def _parse_simple_header(self, reader: Reader) -> Header:
        return SimpleHeader()

    def _parse_multi_header(self, reader: Reader) -> Header:
        # Source format
        id = reader.read_ulong()
        total = reader.read_byte()
        current = reader.read_byte()
        size = reader.read_ushort()  # Some games omit this field!

        if id >> 15:
            # Our crc32() function returns an unsigned int, so we need ulong here.
            decompressed_size = reader.read_ulong()
            checksum = reader.read_ulong()
            compressed = Compression(size=decompressed_size, checksum=checksum)
        else:
            compressed = None

        return MultiHeader(
            id=id,
            current=current,
            total=total,
            size=size,
            compressed=compressed,
        )

    def _parse_payload(self, reader: Reader, header: Header) -> bytes | None:
        if isinstance(header, SimpleHeader):
            return reader.read()
        elif isinstance(header, MultiHeader):
            return self._parse_multi_payload(reader, header)
        elif isinstance(header, MultiGoldsourceHeader):
            return self._parse_multi_payload(reader, header)
        else:
            raise ValueError(f"Unknown header type {header!r}")

    def _parse_multi_payload(
        self,
        reader: Reader,
        header: MultiAnyHeader,
    ) -> bytes | None:
        response = self._get_response(header)
        return self._update_response(reader, header, response)

    def _get_response(
        self,
        header: MultiAnyHeader,
    ) -> MultiPartResponse:
        response = self._responses.get(header.id)
        if response is not None:
            return response

        compressed = header.compressed if isinstance(header, MultiHeader) else None
        response = MultiPartResponse(
            id=header.id,
            current=header.current,
            total=header.total,
            compressed=compressed,
        )
        self._responses[header.id] = response
        return response

    def _update_response(
        self,
        reader: Reader,
        header: MultiAnyHeader,
        response: MultiPartResponse,
    ) -> bytes | None:
        payload = reader.read()

        try:
            response.add(header, payload)
        except ValueError:
            self.invalidate_response(header.id)
            raise

        payload = response.assemble()
        if payload is not None:
            self.invalidate_response(header.id)
            return payload

    def _handle_payload(self, data: bytes) -> None:
        reader = Reader(data)

        header = reader.read_byte()
        if header == 0x41:  # S2C_CHALLENGE
            event = ClientEventChallenge.from_reader(reader)
            self.challenge = event.challenge
        elif header == 0x49:  # A2S_INFO
            event = ClientEventInfo.from_reader(reader)
        elif header == 0x6D:  # Goldsource A2S_INFO
            event = ClientEventGoldsourceInfo.from_reader(reader)
        elif header == 0x44:  # A2S_PLAYER
            event = ClientEventPlayers.from_reader(reader)
        elif header == 0x45:  # A2S_RULES
            event = ClientEventRules.from_reader(reader)
        else:
            raise ValueError(f"Unknown payload header type {header:X}")

        self._events.append(event)


class A2SClientGoldsourceProtocol(A2SClientProtocol):
    """The Goldsource variant of the A2S client protocol used by older games."""

    def _parse_multi_header(self, reader: Reader) -> Header:
        id = reader.read_ulong()
        number = reader.read_byte()
        current = number & 0xF0
        total = number & 0x0F
        return MultiGoldsourceHeader(id=id, current=current, total=total)



from dataclasses import dataclass
from enum import IntEnum
from typing import Self

from little_a2s.enums import _EnumReprMixin
from little_a2s.reader import Reader


# ClientEventInfo types
class ServerType(_EnumReprMixin, IntEnum):
    DEDICATED = ord("d")
    LISTEN = ord("l")
    RELAY = ord("p")


class Environment(_EnumReprMixin, IntEnum):
    LINUX = ord("l")
    WINDOWS = ord("w")
    MACOS_M = ord("m")  # Ergh, can this be combined?
    MACOS_O = ord("o")


class Visibility(_EnumReprMixin, IntEnum):
    PUBLIC = 0
    PRIVATE = 1


class VAC(_EnumReprMixin, IntEnum):
    INSECURE = 0
    SECURE = 1


@dataclass(kw_only=True)
class ExtraInfo:
    """Extra data included with an A2S_INFO response."""

    port: int | None = None
    steam_id: int | None = None
    spectator_port: int | None = None
    spectator_name: str | None = None
    keywords: str | None = None
    game_id: int | None = None

    @classmethod
    def from_reader(cls, reader: Reader, flag: int) -> Self:
        extra = cls()
        if flag & 0x80:
            extra.port = reader.read_ushort()
        if flag & 0x10:
            extra.steam_id = reader.read_uint64()
        if flag & 0x40:
            extra.spectator_port = reader.read_ushort()
            extra.spectator_name = reader.read_null_utf8()
        if flag & 0x20:
            extra.keywords = reader.read_null_utf8()
        if flag & 0x01:
            extra.game_id = reader.read_uint64()
        return extra


# ClientEventGoldsourceInfo types
class GoldsourceModType(_EnumReprMixin, IntEnum):
    SINGLE_AND_MULTIPLAYER = 0
    MULTIPLAYER_ONLY = 1


class GoldsourceModDLL(_EnumReprMixin, IntEnum):
    NATIVE = 0
    """This mod uses the Half-Life DLL."""

    EXTENSION = 1
    """This mod provides its own DLL."""


@dataclass(kw_only=True)
class GoldsourceMod:
    """Extra data included with an A2S_INFO response."""

    link: str
    download_link: str
    version: int
    size: int
    type: GoldsourceModType
    dll: GoldsourceModDLL

    @classmethod
    def from_reader(cls, reader: Reader) -> Self:
        link = reader.read_null_utf8()
        download_link = reader.read_null_utf8()
        reader.read_null()
        version = reader.read_ulong()
        size = reader.read_ulong()
        type = GoldsourceModType(reader.read_byte())
        dll = GoldsourceModDLL(reader.read_byte())

        return cls(
            link=link,
            download_link=download_link,
            version=version,
            size=size,
            type=type,
            dll=dll,
        )


# ClientEventPlayer types
@dataclass(kw_only=True)
class Player:
    """A player returned in the A2S_PLAYER response."""

    index: int
    name: str
    score: int
    duration: float


# Event types
@dataclass(kw_only=True)
class Event:
    """The base class for all A2S protocol events."""


@dataclass(kw_only=True)
class ClientEvent(Event):
    """An A2S client protocol event."""


@dataclass(kw_only=True)
class ClientEventInfo(ClientEvent):
    """An A2S_INFO client protocol event.

    This follows the Source format. For the Goldsource equivalent,
    see :class:`ClientEventGoldsourceInfo`.

    """

    protocol: int
    name: str
    map: str
    folder: str
    game: str
    id: int
    players: int
    max_players: int
    bots: int
    type: ServerType
    environment: Environment
    visibility: Visibility
    vac: VAC
    version: str
    extra: ExtraInfo | None

    @classmethod
    def from_reader(cls, reader: Reader) -> Self:
        protocol = reader.read_byte()
        name = reader.read_null_utf8()
        map = reader.read_null_utf8()
        folder = reader.read_null_utf8()
        game = reader.read_null_utf8()
        id = reader.read_ushort()
        players = reader.read_byte()
        max_players = reader.read_byte()
        bots = reader.read_byte()
        type = ServerType(reader.read_byte())
        environment = Environment(reader.read_byte())
        visibility = Visibility(reader.read_byte())
        vac = VAC(reader.read_byte())
        # Extra data will be here for The Ship
        version = reader.read_null_utf8()

        try:
            extra_flag = reader.read_byte()
        except EOFError:
            extra = None
        else:
            extra = ExtraInfo.from_reader(reader, extra_flag)

        return cls(
            protocol=protocol,
            name=name,
            map=map,
            folder=folder,
            game=game,
            id=id,
            players=players,
            max_players=max_players,
            bots=bots,
            type=type,
            environment=environment,
            visibility=visibility,
            vac=vac,
            version=version,
            extra=extra,
        )


@dataclass(kw_only=True)
class ClientEventGoldsourceInfo(ClientEvent):
    """An A2S_INFO Goldsource client protocol event."""

    address: str
    name: str
    map: str
    folder: str
    game: str
    players: int
    max_players: int
    protocol: int
    type: ServerType
    environment: Environment
    visibility: Visibility
    mod: GoldsourceMod | None
    vac: VAC
    bots: int

    @classmethod
    def from_reader(cls, reader: Reader) -> Self:
        address = reader.read_null_utf8()
        name = reader.read_null_utf8()
        map = reader.read_null_utf8()
        folder = reader.read_null_utf8()
        game = reader.read_null_utf8()
        players = reader.read_byte()
        max_players = reader.read_byte()
        protocol = reader.read_byte()
        type = ServerType(ord(reader.read_char().lower()))
        environment = Environment(ord(reader.read_char().lower()))
        visibility = Visibility(reader.read_byte())
        mod = GoldsourceMod.from_reader(reader) if reader.read_byte() == 1 else None
        vac = VAC(reader.read_byte())
        bots = reader.read_byte()

        return cls(
            address=address,
            name=name,
            map=map,
            folder=folder,
            game=game,
            players=players,
            max_players=max_players,
            protocol=protocol,
            type=type,
            environment=environment,
            visibility=visibility,
            mod=mod,
            vac=vac,
            bots=bots,
        )


@dataclass(kw_only=True)
class ClientEventPlayers(ClientEvent):
    """An A2S_PLAYER client protocol event."""

    players: list[Player]

    @classmethod
    def from_reader(cls, reader: Reader) -> Self:
        players = []
        for _ in range(reader.read_byte()):
            index = reader.read_byte()
            name = reader.read_null_utf8()
            score = reader.read_long()
            duration = reader.read_float()
            # Extra data will be here for The Ship
            player = Player(index=index, name=name, score=score, duration=duration)
            players.append(player)

        return cls(players=players)


@dataclass(kw_only=True)
class ClientEventRules(ClientEvent):
    """An A2S_RULES client protocol event."""

    # While documented to be strings, some games might provide binary data
    # *cough Arma* which may not decode correctly as UTF-8.
    rules: dict[bytes, bytes]

    def decode(self) -> dict[str, str]:
        """Return all rules decoded in UTF-8."""
        return {k.decode(): v.decode() for k, v in self.rules.items()}

    @classmethod
    def from_reader(cls, reader: Reader) -> Self:
        rules = {}
        for _ in range(reader.read_byte()):
            name = reader.read_null_string()
            value = reader.read_null_string()
            rules[name] = value

        return cls(rules=rules)


@dataclass(kw_only=True)
class ClientEventChallenge(ClientEvent):
    """An S2C_CHALLENGE client protocol event."""

    challenge: int

    @classmethod
    def from_reader(cls, reader: Reader) -> Self:
        return cls(challenge=reader.read_long())



# https://developer.valvesoftware.com/wiki/Server_queries
from dataclasses import dataclass
from enum import IntEnum

from little_a2s.enums import _EnumReprMixin


class HeaderType(_EnumReprMixin, IntEnum):
    SIMPLE = -1
    MULTI = -2


@dataclass(kw_only=True)
class Compression:
    """The compression header for Source games. Mostly present in ~2006-era engines."""

    size: int
    checksum: int


@dataclass(kw_only=True)
class Header:
    """An A2S response header."""


@dataclass(kw_only=True)
class SimpleHeader(Header):
    """An single-packet A2S response header."""


@dataclass(kw_only=True)
class MultiHeader(Header):
    """A multi-packet A2S response header for Source games."""

    id: int
    current: int
    total: int
    size: int  # Some games omit this field!
    compressed: Compression | None


@dataclass(kw_only=True)
class MultiGoldsourceHeader(Header):
    """A multi-packet A2S response header for Goldsource games."""

    id: int
    current: int
    total: int



from dataclasses import dataclass, field


@dataclass(kw_only=True)
class Packet:
    """The base class of all A2S packets."""


@dataclass(kw_only=True)
class ClientPacket(Packet):
    """An A2S packet sent by the client."""

    header: bytes
    payload: bytes
    challenge: int

    def __bytes__(self) -> bytes:
        challenge = self.challenge.to_bytes(4, "little", signed=True)
        return b"".join((self.header, self.payload, challenge))


@dataclass(kw_only=True)
class ClientPacketInfo(ClientPacket):
    """An A2S_INFO packet sent by the client."""

    header: bytes = field(default=b"\xff\xff\xff\xff\x54", init=False)
    payload: bytes = field(default=b"Source Engine Query\x00", init=False)
    challenge: int = -1


@dataclass(kw_only=True)
class ClientPacketPlayers(ClientPacket):
    """An A2S_PLAYER packet sent by the client."""

    header: bytes = field(default=b"\xff\xff\xff\xff\x55", init=False)
    payload: bytes = field(default=b"", init=False)
    challenge: int = -1


@dataclass(kw_only=True)
class ClientPacketRules(ClientPacket):
    """An A2S_RULES packet sent by the client."""

    header: bytes = field(default=b"\xff\xff\xff\xff\x56", init=False)
    payload: bytes = field(default=b"", init=False)
    challenge: int = -1



import struct
from io import BytesIO
from typing import Literal, Protocol, runtime_checkable


@runtime_checkable
class Readable(Protocol):
    def read(self, n: int = -1, /) -> bytes: ...


class Reader:
    """A simple reader for parsing serialized data."""

    def __init__(self, data: bytes | Readable) -> None:
        if not isinstance(data, Readable):
            data = BytesIO(data)

        self._file = data

    def read(self, n: int = -1, /) -> bytes:
        data = self._file.read(n)
        if n >= 0 and len(data) < n:
            raise EOFError
        return data

    def read_byte(self) -> int:
        return self.read(1)[0]

    def read_char(self) -> str:
        return self.read(1).decode("ascii")

    def read_null(self) -> Literal[0]:
        n = self.read_byte()
        if n != 0:
            raise ValueError(f"Expected null byte, got {n!r}")
        return n

    def read_short(self) -> int:
        return int.from_bytes(self.read(2), "little", signed=True)

    def read_ushort(self) -> int:
        return int.from_bytes(self.read(2), "little")

    def read_long(self) -> int:
        return int.from_bytes(self.read(4), "little", signed=True)

    def read_ulong(self) -> int:
        return int.from_bytes(self.read(4), "little")

    def read_float(self) -> float:
        return struct.unpack("<f", self.read(4))[0]

    def read_uint64(self) -> int:
        return int.from_bytes(self.read(8), "little")

    def read_null_string(self) -> bytes:
        return self.read_until(0)

    def read_null_utf8(self) -> str:
        s = self.read_null_string()
        return s.decode()

    def read_until(self, sep: int) -> bytes:
        """Read until the sep character is found and return all bytes before sep."""
        if not 0x00 <= sep < 0xFF:
            raise ValueError(f"Expected sep in range [0, 255], got {sep!r}")

        # FIXME: is this hot loop slow?
        data = bytearray()
        while (char := self.read_byte()) != sep:
            data.append(char)

        return bytes(data)

    def read_varchar1(self) -> bytes:
        length = self.read_byte()
        return self.read(length)

    def read_varchar2(self) -> bytes:
        length = self.read_ushort()
        return self.read(length)

    def read_varchar4(self) -> bytes:
        length = self.read_ulong()
        return self.read(length)

    def read_varchar8(self) -> bytes:
        length = self.read_uint64()
        return self.read(length)



from .client import (
    A2SClientProtocol as A2SClientProtocol,
    A2SClientGoldsourceProtocol as A2SClientGoldsourceProtocol,
    MultiPartResponse as MultiPartResponse,
)
from .events import (
    ClientEvent as ClientEvent,
    ClientEventChallenge as ClientEventChallenge,
    ClientEventGoldsourceInfo as ClientEventGoldsourceInfo,
    ClientEventInfo as ClientEventInfo,
    ClientEventPlayers as ClientEventPlayers,
    ClientEventRules as ClientEventRules,
    Environment as Environment,
    Event as Event,
    ExtraInfo as ExtraInfo,
    GoldsourceMod as GoldsourceMod,
    GoldsourceModDLL as GoldsourceModDLL,
    GoldsourceModType as GoldsourceModType,
    Player as Player,
    ServerType as ServerType,
    VAC as VAC,
    Visibility as Visibility,
)
from .headers import (
    Compression as Compression,
    Header as Header,
    HeaderType as HeaderType,
    MultiGoldsourceHeader as MultiGoldsourceHeader,
    MultiHeader as MultiHeader,
    SimpleHeader as SimpleHeader,
)
from .packets import (
    ClientPacket as ClientPacket,
    ClientPacketInfo as ClientPacketInfo,
    ClientPacketPlayers as ClientPacketPlayers,
    ClientPacketRules as ClientPacketRules,
    Packet as Packet,
)
from .reader import Readable as Readable, Reader as Reader
